akprojectileHandler = {
	shellNum = 1,
	shells = {},
	defaultShell = {
		active = false,
		aliveTimer = 0
	},
}

akshellHandler = {
	shellNum = 1,
	shells = {},
	defaultShell = {
		active = false,
		aliveTimer = 0,
		bounces = 0
	},
}

function init()
	RegisterTool("ak", "VSV Metro 2033", "MOD/vox/ak.vox", 3)
	SetBool("game.tool.ak.enabled", true)
	SetString("game.tool.ak.ammo.display","")
	SetFloat("game.tool.ak.ammo", 101)
	realisticdamage = GetBool("savegame.mod.realisticdamage")

	damage = realisticdamage and 0.25 or 0.6
	gravity = Vec(0, -1, 0)
	shellGravity = Vec(0, -50, 0)
	velocity = 1.5

	gunsound = LoadSound("MOD/snd/ak0.ogg")
	cocksound = LoadSound("MOD/snd/guncock.ogg")
	reloadsound = LoadSound("MOD/snd/reload.ogg")
	dryfiresound = LoadSound("MOD/snd/dryfire.ogg")
	refillsound = LoadSound("MOD/snd/refill.ogg")
	shellsound = LoadSound("MOD/snd/shellground.ogg")

	reloadTime = 1.10
	shotDelay = 0.085
	spreadTimer = 0
	ammo = 20
	mags = 10
	reloading = false
	ironsight = false
	unlimitedammo = GetBool("savegame.mod.unlimitedammo")

	for i=1, 200 do
		akprojectileHandler.shells[i] = deepcopy(akprojectileHandler.defaultShell)
	end

	for i=1, 100 do
		akshellHandler.shells[i] = deepcopy(akshellHandler.defaultShell)
	end

	shootTimer = 0
	reloadTimer = 0
	recoilTimer = 0
	lightTimer = 0

	magoutTimer = 0
	maginTimer = 0
end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end

function GetAimPos()
	local ct = GetCameraTransform()
	local forwardPos = TransformToParentPoint(ct, Vec(0, 0, -100))
    local direction = VecSub(forwardPos, ct.pos)
    local distance = VecLength(direction)
	local direction = VecNormalize(direction)
	local hit, hitDistance = QueryRaycast(ct.pos, direction, distance)
	if hit then
		forwardPos = TransformToParentPoint(ct, Vec(0, 0, -hitDistance))
		distance = hitDistance
	end
	return forwardPos, distance
end

function rndVec(length)
	local v = VecNormalize(Vec(math.random(-100,100), math.random(-100,100), math.random(-100,100)))
	return VecScale(v, length)	
end

function Shoot()
	if shootTimer > 0 or ammo == 0 then
		return
	end

	aimpos, distance = GetAimPos()

	local ct = GetCameraTransform()
	local fwdpos = TransformToParentPoint(ct, Vec(0, 0, -100))
	local dir = VecSub(aimpos, toolPos)
	local straightDir = VecSub(fwdpos, toolPos)

	local maxSpread = InputDown("ctrl") and 1 or 2
	if ironsight then maxSpread = maxSpread / 2 end
	local spread = math.min(spreadTimer, maxSpread) * distance/100
	dir[1] = dir[1] + (math.random()-0.5)*2*spread
	dir[2] = dir[2] + (math.random()-0.5)*2*spread
	dir[3] = dir[3] + (math.random()-0.5)*2*spread

	local projectile = Spawn("MOD/akbullet.xml", Transform(toolPos, QuatLookAt(dir)))
	
	local espread = 0.5
	local ejectPos = TransformToParentPoint(toolTrans, Vec(0.2105, -0.65, -0.10))
	local ejectDirVec = TransformToParentPoint(toolTrans, Vec(1.5, 3, -0.9))
	local shell = Spawn("MOD/akshell.xml", Transform(ejectPos, QuatLookAt(straightDir)))
	ejectDirVec[1] = ejectDirVec[1] + (math.random()-0.5)*2*espread
	ejectDirVec[2] = ejectDirVec[2] + (math.random()-0.5)*2*espread
	ejectDirVec[3] = ejectDirVec[3] + (math.random()-0.5)*2*espread
	local ejectDir = VecSub(ejectDirVec, ejectPos)

	straightDir[1] = straightDir[1] + (math.random()-0.5)*10*espread
	straightDir[2] = straightDir[2] + (math.random()-0.5)*10*espread
	straightDir[3] = straightDir[3] + (math.random()-0.5)*10*espread

	akshellHandler.shells[akshellHandler.shellNum] = deepcopy(akshellHandler.defaultShell)
	loadedShell = akshellHandler.shells[akshellHandler.shellNum] 
	loadedShell.active = true
	loadedShell.pos = ejectPos
	loadedShell.targetRot = QuatLookAt(straightDir)
	loadedShell.voxBody = shell[1]
	loadedShell.aliveTimer = 3
	loadedShell.predictedBulletVelocity = VecScale(ejectDir, 2)
	akshellHandler.shellNum = (akshellHandler.shellNum%#akshellHandler.shells) + 1

	akprojectileHandler.shells[akprojectileHandler.shellNum] = deepcopy(akprojectileHandler.defaultShell)
	loadedShell = akprojectileHandler.shells[akprojectileHandler.shellNum] 
	loadedShell.active = true
	loadedShell.pos = toolPos
	loadedShell.voxBody = projectile[1]
	loadedShell.aliveTimer = 3
	loadedShell.predictedBulletVelocity = VecScale(dir, velocity*(100/distance))
	akprojectileHandler.shellNum = (akprojectileHandler.shellNum%#akprojectileHandler.shells) + 1

	SpawnParticle("fire", toolPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), 0.3, 0.1)
	PlaySound(gunsound, GetPlayerTransform().pos, 1, false)

	if not unlimitedammo then
		ammo = ammo - 1
	end
	shootTimer = shotDelay
	recoilTimer = shotDelay
	lightTimer = shotDelay/2
	spreadTimer = spreadTimer + 1.25
end

function destructiblerobots(pos, damage, hitcounter)
    local hit, point, n, shape = QueryClosestPoint(pos, 0.3)
    if hit then 
		SetFloat('level.destructible-bot.debug', false)
        SetInt('level.destructible-bot.hitCounter', hitcounter)
        SetInt('level.destructible-bot.hitShape', shape)
        SetString('level.destructible-bot.weapon', "AK-")
        SetFloat('level.destructible-bot.damage', damage)
    end 
end

function ProjectileOperations(projectile)
	projectile.predictedBulletVelocity = VecAdd(projectile.predictedBulletVelocity, (VecScale(gravity, GetTimeStep())))
	local point2 = VecAdd(projectile.pos, VecScale(projectile.predictedBulletVelocity, GetTimeStep()))
	local dir = VecNormalize(VecSub(point2, projectile.pos))
	QueryRejectBody(projectile.voxBody)
	QueryRequire("physical")
	local hit, dist = QueryRaycast(projectile.pos, dir, VecLength(VecSub(point2, projectile.pos)))
	
	if hit then
		hitPos = VecAdd(projectile.pos, VecScale(VecNormalize(VecSub(point2, projectile.pos)), dist))
		SetBodyTransform(projectile.voxBody, Transform(projectile.pos, QuatLookAt(projectile.pos, point2)))
		SetBodyDynamic(projectile.voxBody, true)
		destructiblerobots(hitPos, 10, 1)
		projectile.active = false
		PlaySound(bouncesound, projectile.pos, 1, false)
		MakeHole(hitPos, damage, damage*0.85, damage*0.10)
		SpawnParticle("smoke", hitPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), damage, 1)
	else
	end
	SetBodyTransform(projectile.voxBody, Transform(projectile.pos, QuatLookAt(projectile.pos, point2)))
	projectile.pos = point2
end

function ShellOperations(shell)
	shell.predictedBulletVelocity = VecAdd(shell.predictedBulletVelocity, (VecScale(shellGravity, GetTimeStep())))
	local point2 = VecAdd(shell.pos, VecScale(shell.predictedBulletVelocity, GetTimeStep()))
	local dir = VecNormalize(VecSub(point2, shell.pos))
	local distance = VecLength(VecSub(point2, shell.pos))
	QueryRejectBody(shell.voxBody)
	QueryRequire("physical")
	local hit, dist, normal = QueryRaycast(shell.pos, dir, distance)
	if hit then
		if shell.bounces == 5 then
			shell.active = false
		elseif shell.bounces == 0 then
			shell.targetRot = QuatLookAt(shell.pos, point2)
			PlaySound(shellsound, shell.pos, 0.6, false)
			local dot = VecDot(normal, shell.predictedBulletVelocity)
			shell.predictedBulletVelocity = VecSub(shell.predictedBulletVelocity, VecScale(normal, dot*1.6))
			shell.bounces = shell.bounces + 1
		else
			shell.targetRot = QuatLookAt(point2, shell.pos)
			local dot = VecDot(normal, shell.predictedBulletVelocity)
			shell.predictedBulletVelocity = VecSub(shell.predictedBulletVelocity, VecScale(normal, dot*1.6))
			shell.bounces = shell.bounces + 1
		end
	end
	shell.rot = QuatSlerp(shell.rot, shell.targetRot, 0.3)
	SetBodyTransform(shell.voxBody, Transform(shell.pos, shell.rot))
	shell.pos = point2
end

function Reload()
	if reloading then
		return
	end
	reloading = true
	PlaySound(reloadsound, GetPlayerTransform().pos, 0.6, false)
	reloadTimer = reloadTime
	magoutTimer = 0.6
	mags = mags - 1
end

function tick(dt)
	if GetString("game.player.tool") == "ak" and GetBool("game.player.canusetool") then
		if InputDown("usetool") and ammo > 0 and not reloading then
			Shoot()
		end

		if InputPressed("usetool") and not reloading then
			spreadTimer = 0
			if ammo == 0 then
				PlaySound(dryfiresound, GetPlayerTransform().pos, 1, false)
			end
		end

		if InputReleased("usetool") and ammo > 0 then
			SpawnParticle("darksmoke", toolPos, Vec(0, 1.0+math.random(1,10)*0.1, 0), 0.3, 0.5)
		end

		if InputPressed("rmb") and not IsHandleValid(GetPlayerGrabBody()) then
			ironsight = not ironsight
		end

		local b = GetToolBody()
		if b ~= 0 then
			local heightOffset = InputDown("ctrl") and 0.3 or 0.2
			local magoffset = Vec(0, 0, 0)
			local magtimer = magoutTimer + maginTimer
			local offset = Transform(Vec(0, heightOffset, 0))
			local x, y, z, rot = 0, heightOffset, 0, 0
			if ironsight then
				x = 0.2105
				y = 0.45
				z = 1
				rot = 2.5
				offset = Transform(Vec(-x, y, z), QuatEuler(-rot, 0, 0))
			end

			if magtimer > 0 then
				offset.rot = QuatEuler(10, 0, -10)
				offset.pos = VecAdd(offset.pos, Vec(0.2, 0.2, 0))
				magoffset = Vec(-0.6, -0.6, 0.6)
			end

			SetToolTransform(offset, ironsight and 0 or 0.5)
			toolTrans = GetBodyTransform(b)
			toolPos = TransformToParentPoint(toolTrans, Vec(0.2105, -0.6, -2.6))

			if recoilTimer > 0 then
				local t = Transform()
				t.pos = Vec(-x, y, recoilTimer+z)
				ironrot = ironsight and rot or -recoilTimer*50-rot
				t.rot = QuatEuler(-ironrot, 0, 0)
				SetToolTransform(t)

				recoilTimer = recoilTimer - dt
			end

			if lightTimer > 0 then
				PointLight(toolPos, 1, 1, 1, 0.5)
				lightTimer = lightTimer - dt
			end
			
			if magoutTimer > 0 then
				magoffset = Vec(-0.3+magoutTimer/2, -0.6+magoutTimer, 0.9-magoutTimer*1.5)
				magoutTimer = magoutTimer - dt
				if magoutTimer < 0 then
					maginTimer = 0.6
				end
			end

			if maginTimer > 0 then
				magoffset = Vec(-maginTimer/2, -maginTimer, maginTimer*1.5)
				maginTimer = maginTimer - dt
			end
			
			if body ~= b then
				body = b
				local shapes = GetBodyShapes(b)
				mag = shapes[2]
				magTrans = GetShapeLocalTransform(mag)
			end

			mt = TransformCopy(magTrans)
			mt.pos = VecAdd(mt.pos, magoffset)
			mt.rot = QuatRotateQuat(mt.rot, QuatEuler(-magtimer*20, magtimer*20, 0))
			SetShapeLocalTransform(mag, mt)
		end

		if not unlimitedammo then
			if ammo < 20 and mags > 1 and InputPressed("R") then
				Reload()
			end
			
			if GetBool("ammobox.refill") then
				SetBool("ammobox.refill", false)
				mags = mags + 1
				PlaySound(refillsound, GetPlayerTransform().pos, 1, false)
			end

			if reloading then
				ironsight = false
				reloadTimer = reloadTimer - dt
				if reloadTimer < 0 then
					ammo = 20
					reloadTimer = 0
					PlaySound(cocksound, GetPlayerTransform().pos, 0.85)
					reloading = false
				end
			end
		end

		for key, shell in ipairs(akprojectileHandler.shells) do
			if shell.aliveTimer > 0 then
				shell.aliveTimer = shell.aliveTimer - dt
				if shell.aliveTimer < 0.1 then
					shell.active = false
					shell.aliveTimer = 0
				end
			end
			if shell.active then
				ProjectileOperations(shell)
			else
				Delete(shell.voxBody)
			end
		end

		for key, shell in ipairs(akshellHandler.shells) do
			if shell.aliveTimer > 0 then
				shell.aliveTimer = shell.aliveTimer - dt
				if shell.aliveTimer < 0.1 then
					shell.active = false
				end
			end
			if shell.active then
				ShellOperations(shell)
			else
				Delete(shell.voxBody)
			end
		end
	
		if shootTimer > 0 or ammo == 0 then
			shootTimer = shootTimer - dt
		end
	else
		ironsight = false
	end
end

function draw()
	if GetString("game.player.tool") == "ak" and GetPlayerVehicle() == 0 then
		if not unlimitedammo then
			UiPush()
				UiTranslate(UiCenter(), UiHeight()-60)
				UiAlign("center middle")
				UiColor(1, 1, 1)
				UiFont("bold.ttf", 32)
				UiTextOutline(0,0,0,1,0.1)
				if reloading then
					UiText("Reloading")
				else
					UiText(ammo.."/"..20*math.max(0, mags-1))
				end
			UiPop()
		end
	end
end